// Copyright 2000-2023 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
package com.jetbrains.python.sdk.add

import com.intellij.execution.target.TargetEnvironmentConfiguration
import com.intellij.execution.wsl.WSLDistribution
import com.intellij.execution.wsl.WslPath.Companion.parseWindowsUncPath
import com.intellij.execution.wsl.target.WslTargetEnvironmentConfiguration
import com.intellij.util.SystemProperties
import com.intellij.util.text.nullize
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import java.nio.file.Path

class ProjectLocationContexts {
  fun getProjectLocationContextFor(projectPath: String): ProjectLocationContext {
    val wslPath = parseWindowsUncPath(projectPath)
    return if (wslPath != null) WslContext(wslPath.distribution) else LocalContext
  }
}

interface ProjectLocationContext {
  val targetEnvironmentConfiguration: TargetEnvironmentConfiguration?

  suspend fun fetchUserHomeDirectory(): Path?
}

data object LocalContext : ProjectLocationContext {
  override val targetEnvironmentConfiguration: TargetEnvironmentConfiguration? = null

  override suspend fun fetchUserHomeDirectory(): Path? = SystemProperties.getUserHome().nullize(nullizeSpaces = true)?.let { Path.of(it) }
}

data class WslContext(val distribution: WSLDistribution) : ProjectLocationContext {
  override val targetEnvironmentConfiguration: TargetEnvironmentConfiguration = WslTargetEnvironmentConfiguration(distribution)

  override suspend fun fetchUserHomeDirectory(): Path? =
    withContext(Dispatchers.IO) {
      val userHome = distribution.userHome
      userHome?.let { distribution.getWindowsPath(userHome) }?.let { Path.of(it) }
    }
}